/*=========================================================================

  Program:   Insight Segmentation & Registration Toolkit
  Module:    $RCSfile: itkUnicodeIOTest.cxx,v $
  Language:  C++
  Date:      $Date: 2009-10-27 20:43:13 $xgoto-l

  Version:   $Revision: 1.0 $

  Copyright (c) 2002 Insight Consortium. All rights reserved.
  See ITKCopyright.txt or http://www.itk.org/HTML/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notices for more information.

=========================================================================*/
#include "utf8.h"

#include <cassert>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <string>
#include <vector>

#if defined(WIN32) || defined(_WIN32) || defined(__CYGWIN__)
#  include <Windows.h>
#  include <Winnls.h>
#endif

// Cygwin has no wstring by default but seems to have
// wchar_t... This typedef just helps compiling
// but does not make things work. cygwin does not provide
// _wfopen so we are doomed anyhow
#ifdef __CYGWIN__
namespace std
{
  typedef std::basic_string <wchar_t> wstring;
}
#endif

// Get the size of a wchar_t
#ifdef __SIZEOF_WCHAR_T__
#  define LOCAL_SIZEOF_WCHAR_T __SIZEOF_WCHAR_T__
#elif defined(__WCHAR_MAX__)
#  if (__WCHAR_MAX__ == 2147483647)
#    define LOCAL_SIZEOF_WCHAR_T 4
#  elif  (__WCHAR_MAX__ == 65535)
#    define LOCAL_SIZEOF_WCHAR_T 2
#  elif  (__WCHAR_MAX__ == 255)
#    define LOCAL_SIZEOF_WCHAR_T 1
#  else
#    define LOCAL_SIZEOF_WCHAR_T 0
#  endif
//#elif defined(_MSC_VER)
#elif defined(WIN32) || defined(_WIN32) || defined(__CYGWIN__)
#  define LOCAL_SIZEOF_WCHAR_T 2
#else
#  define LOCAL_SIZEOF_WCHAR_T 0
#endif


namespace itk
{
namespace Utf8
{

// Check if the string is really encoded in utf-8
inline bool IsValidUtf8(const std::string & str)
{
  return utf8::is_valid(str.begin(), str.end());
}

#if defined(WIN32) || defined(_WIN32) || defined(__CYGWIN__)
// Check if the string is really encoded in utf-8 using windows API
// (available also on cygwin and mingw). Apparently this does not actually work
// I just put it here for reference on why utfcpp is useful
inline bool IsValidUtf8_MSVC(const std::string & str)
{
  // According to MSDN, MultiByteToWideChar should return 0 if there
  // was a problem during conversion
  const int utf16_size = MultiByteToWideChar(CP_UTF8, 0, str.c_str(), str.length(), 0, 0);
  return (utf16_size != 0);
}
#endif

inline std::wstring Utf8StringToWString( const std::string & str )
{
  std::wstring wstr;
#if (LOCAL_SIZEOF_WCHAR_T == 4)
  utf8::utf8to32(str.begin(), str.end(), std::back_inserter(wstr));
#elif (LOCAL_SIZEOF_WCHAR_T == 2)
  utf8::utf8to16(str.begin(), str.end(), std::back_inserter(wstr));
#elif (LOCAL_SIZEOF_WCHAR_T == 1)
  wstr = str;
#else
  assert(false);
#endif
  return wstr;
}

inline std::string WStringToUtf8String( const std::wstring & wstr )
{
  std::string str;
#if (LOCAL_SIZEOF_WCHAR_T == 4)
  utf8::utf32to8(wstr.begin(), wstr.end(), std::back_inserter(str));
#elif (LOCAL_SIZEOF_WCHAR_T == 2)
  utf8::utf16to8(wstr.begin(), wstr.end(), std::back_inserter(str));
#elif (LOCAL_SIZEOF_WCHAR_T == 1)
  str = wstr;
#else
  assert(false);
#endif
  return str;
}

inline FILE * utf8fopen ( const std::string & str, const std::string & mode )
{
#if (defined(WIN32) || defined(_WIN32)) && (!defined(__CYGWIN__))
  // cygwin has no _wfopen but mingw has
  // Convert to utf16
  const std::wstring str_utf16 = Utf8StringToWString( str );
  const std::wstring mode_utf16 = Utf8StringToWString( mode );
  return _wfopen(str_utf16.c_str(), mode_utf16.c_str());
#else
  return fopen(str.c_str(), mode.c_str());
#endif
}

std::ifstream * createReadStream(const char * filename)
{
  std::ifstream * readStream = new std::ifstream;
#if (defined(WIN32) || defined(_WIN32)) && (!(defined(__CYGWIN__) || defined(__MINGW32__)))
  // Neither cygwin nor mingw have a ifstream constructor that takes a wchar_t * filename
  const std::wstring filename_utf16 = Utf8StringToWString( filename );
  readStream->open(filename_utf16.c_str(), std::ios::binary | std::ios::in);
#elif defined(__sgi)
  readStream->open(filename, std::ios::in);
#else
  readStream->open(filename, std::ios::binary | std::ios::in);
#endif
  return readStream;
}

std::ofstream * createWriteStream(const char * filename, const bool append)
{
  std::ofstream * writeStream = new std::ofstream;
#if (defined(WIN32) || defined(_WIN32)) && (!(defined(__CYGWIN__) || defined(__MINGW32__)))
  // Neither cygwin nor mingw have a ofstream constructor that takes a wchar_t * filename
  const std::wstring filename_utf16 = Utf8StringToWString( filename );
  if(!append)
    {
    writeStream->open(filename_utf16.c_str(), std::ios::binary | std::ios::out);
    }
  else
    {
    writeStream->open(filename_utf16.c_str(), std::ios::binary | std::ios::app | std::ios::out);
    }
#elif defined(__sgi)
  // Some older sgi compilers have a error in the ofstream constructor
  // that requires a file to exist for output
  {
  std::ofstream tFile(filename, std::ios::out);
  tFile.close();
  }
  if(!append)
    {
    writeStream->open(filename, std::ios::out);
    }
  else
    {
    writeStream->open(filename, std::ios::app | std::ios::out);
    }
#else
  if(!append)
    {
    writeStream->open(filename, std::ios::binary | std::ios::out);
    }
  else
    {
    writeStream->open(filename, std::ios::binary | std::ios::app |
                                std::ios::out);
    }
#endif
  return writeStream;
}

} // end namespace
} // end namespace

#undef LOCAL_SIZEOF_WCHAR_T



// Some utility functions for the test

// Check if alpha.txt exists using _wfopen with a wstring on MSVC and mingw
// and fopen with a UTF-8 char * otherwise
bool checkAlphaExists()
{
#if (defined(WIN32) || defined(_WIN32)) && (!defined(__CYGWIN__))
  // cygwin has no _wfopen but mingw has
  const std::wstring wstr( L"\u03B1.txt" );
  FILE * tmp = _wfopen(wstr.c_str(), L"r");
#else
  std::string utf8_str;
  utf8_str.push_back(0xCE);
  utf8_str.push_back(0xB1);
  utf8_str += ".txt";
  FILE * tmp = fopen(utf8_str.c_str(), "r");
#endif
  if (tmp!=0)
    {
    // closing is required at least for mingw
    fclose(tmp);
    return true;
    }
  return false;
}

// Try to delete alpha.txt using wunlink with a wstring on MSVC and unlink with UTF-8 char * otherwise
bool removeAlpha()
{
#if (defined(WIN32) || defined(_WIN32)) && (!defined(__CYGWIN__))
  // cygwin has no _wunlink but mingw has
  const std::wstring wstr( L"\u03B1.txt" );
  return (_wunlink(wstr.c_str())!=-1);
#else
  std::string utf8_str;
  utf8_str.push_back(0xCE);
  utf8_str.push_back(0xB1);
  utf8_str += ".txt";
  return (unlink(utf8_str.c_str())!=-1);
#endif
}





int main( int argc, char** argv)
{
  std::cout << "Starting test." << std::endl;
  
  int nberror = 0;
   
  // Put alpha.txt encoded in utf8 within a std::string
  std::string utf8_str;
  utf8_str.push_back(0xCE);
  utf8_str.push_back(0xB1);
  utf8_str += ".txt";

  if ( !itk::Utf8::IsValidUtf8(utf8_str) )
    {
    std::cout << "Wrongly detected invalid utf8 string." << std::endl;
    ++nberror;
    }


#if defined(WIN32) || defined(_WIN32) || defined(__CYGWIN__)
  if ( !itk::Utf8::IsValidUtf8_MSVC(utf8_str) )
    {
    std::cout << "Wrongly detected invalid utf8 string using windows API." << std::endl;
    //++nberror;
    // Do not increment nberror since this is just to show some shortcomings
    // in the windows API method
    }
#endif


#ifndef __CYGWIN__
  // cygwin really does not like unicode
  // The following piece of code compile but removes any console output!
  const std::string utf8_str_backandforth
    = itk::Utf8::WStringToUtf8String(itk::Utf8::Utf8StringToWString(utf8_str));
  std::cout << "utf8_str_backandforth=" << utf8_str_backandforth << std::endl;
  if ( utf8_str_backandforth != utf8_str )
    {
    std::cout << "Conversion from utf8 string to wstring and back failed." << std::endl;
    ++nberror;
    }
#endif
  
  
  // Create a non utf8 std::string
  std::string bad_utf8_str;
  bad_utf8_str.push_back(0xC0);
  bad_utf8_str.push_back(0xC0);
  bad_utf8_str += ".txt";

  if ( itk::Utf8::IsValidUtf8(bad_utf8_str) )
    {
    std::cout << "Did not detect invalid utf8 string." << std::endl;
    ++nberror;
    }

#if defined(WIN32) || defined(_WIN32) || defined(__CYGWIN__)
  if ( itk::Utf8::IsValidUtf8_MSVC(bad_utf8_str) )
    {
    std::cout << "Did not detect invalid utf8 string using windows API." << std::endl;
    //++nberror;
    // Do not increment nberror since this is just to show some shortcomings
    // in the windows API method
    }
#endif



  // Start by removing alpha.txt if it exists
  removeAlpha();

  
  // Create alpha.txt using utf8fopen
  FILE * wfile = itk::Utf8::utf8fopen(utf8_str, "wb");
  
  if (!checkAlphaExists())
    {
    std::cout << "alpha.txt does not exist after utf8fopen." << std::endl;
    ++nberror;
    }

  if (wfile!=NULL)
    {
    fputs("test",wfile);
    fclose(wfile);
    }
  else
    {
    std::cout << "wfile is null." << std::endl;
    ++nberror;
    }

  FILE * rfile = itk::Utf8::utf8fopen(utf8_str, "rb");

  if (rfile!=NULL)
    {
    char teststring[10];
    fgets(teststring, 10, rfile);
    std::cout << "teststring=" << teststring <<std::endl;

    if ( strcmp( teststring, "test" ) != 0 )
      {
      std::cout << "teststring is not equal to test." << std::endl;
      ++nberror;
      }

    fclose(rfile);
    }
  else
    {
    std::cout << "rfile is null." << std::endl;
    ++nberror;
    }

  if (!removeAlpha())
    {
    std::cout << "Could not remove alpha.txt after utf8fopen." << std::endl;
    ++nberror;
    }


  // Create alpha.txt using streams
  std::ofstream * wstream = itk::Utf8::createWriteStream(utf8_str.c_str(), false);
  
  if (!checkAlphaExists())
    {
    std::cout << "alpha.txt does not exist after createWriteStream." << std::endl;
    ++nberror;
    }

  if (wstream!=NULL)
    {
    if ( wstream->is_open() )
      {
      *wstream << "test2";
      wstream->close();
      }
    else
      {
      std::cout << "wstream is not open." << std::endl;
      ++nberror;
      }
    delete wstream;
    }
  else
    {
    std::cout << "wstream is null." << std::endl;
    ++nberror;
    }

  std::ifstream * rstream = itk::Utf8::createReadStream(utf8_str.c_str());

  if (rstream!=NULL)
    {
    std::string teststring2;
    if ( rstream->is_open() )
      {
      std::getline(*rstream, teststring2);
      std::cout << "teststring2=" << teststring2 <<std::endl;

      if ( teststring2 != std::string("test2") )
        {
        std::cout << "teststring2 is not equal to test2." << std::endl;
        ++nberror;
        }

      rstream->close();
      }
    else
      {
      std::cout << "rstream is not open." << std::endl;
      ++nberror;
      }

    delete rstream;
    }
  else
    {
    std::cout << "rstream is null." << std::endl;
    ++nberror;
    }

  if (!removeAlpha())
    {
    std::cout << "Could not remove alpha.txt after stream open." << std::endl;
    ++nberror;
    }
  

  if ( nberror > 0 ) 
    {  
    std::cout << "Test failed with " << nberror << " errors." << std::endl;
    return EXIT_FAILURE;
    }

  std::cout << "Test passed." << std::endl;
  return EXIT_SUCCESS;
}
